from datetime import datetime

from pydantic import BaseModel


class BubbleInfo(BaseModel):
    add: int
    ts: int


class Bubble(BaseModel):
    normal: BubbleInfo
    assist: BubbleInfo


class BuildingChar(BaseModel):
    """基建进驻干员信息"""

    charId: str
    ap: int
    lastApAddTime: int
    index: int
    bubble: Bubble
    workTime: int


class Labor(BaseModel):
    """无人机"""

    maxValue: int
    value: int
    lastUpdateTime: int
    remainSecs: int

    @property
    def labor_now(self) -> int:
        if self.maxValue == self.value:
            return self.maxValue
        elapsed_time = datetime.now().timestamp() - self.lastUpdateTime
        labor_increment = elapsed_time / (self.remainSecs / (self.maxValue - self.value))
        return min(int(labor_increment + self.value), self.maxValue)


class Furniture(BaseModel):
    """家具持有数"""

    total: int



from pydantic import BaseModel

from .base import Equip


class Skill(BaseModel):
    """干员技能"""

    id: str
    specializeLevel: int


class Character(BaseModel):
    """持有干员"""

    charId: str
    skinId: str
    level: int
    evolvePhase: int
    potentialRank: int
    mainSkillLvl: int
    skills: list[Skill]
    equip: list[Equip]
    favorPercent: int
    defaultSkillId: str
    gainTime: int
    defaultEquipId: str



from nonebot.exception import NoneBotException


class Exception(NoneBotException):
    """异常基类"""


class RequestException(Exception):
    """请求错误"""


class UnauthorizedException(Exception):
    """登录授权错误"""


class LoginException(Exception):
    """登录错误"""



from nonebot import logger, get_driver

from .config import config
from .download import GameResourceDownloader

driver = get_driver()

RESOURCE_ROUTES = ["portrait", "avatar", "skill"]


@driver.on_startup
async def startup():
    if config.check_res_update:
        if version := await GameResourceDownloader.check_update():
            logger.info("开始下载游戏资源")
            for route in RESOURCE_ROUTES:
                logger.info(f"正在下载: {route}")
                await GameResourceDownloader.download_all(
                    owner="yuanyan3060", repo="ArknightsGameResource", route=route, branch="main"
                )
            if isinstance(version, str):
                GameResourceDownloader.update_version_file(version)
                logger.success(f"游戏资源已更新到版本：{version}")



import httpx

from ..schemas import CRED
from ..exception import RequestException

app_code = "4ca99fa6b56cc2ba"


class SklandLoginAPI:
    _headers = {
        "User-Agent": ("Skland/1.32.1 (com.hypergryph.skland; build:103201004; Android 33; ) Okhttp/4.11.0"),
        "Accept-Encoding": "gzip",
        "Connection": "close",
    }

    @classmethod
    async def get_grant_code(cls, token: str) -> str:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                "https://as.hypergryph.com/user/oauth2/v2/grant",
                json={"appCode": app_code, "token": token, "type": 0},
                headers={**cls._headers},
            )

            if status := response.json().get("status"):
                if status != 0:
                    raise RequestException(f"使用token获得认证代码失败：{response.json().get('msg')}")
            return response.json()["data"]["code"]

    @classmethod
    async def get_cred(cls, grant_code: str) -> CRED:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                "https://zonai.skland.com/api/v1/user/auth/generate_cred_by_code",
                json={"code": grant_code, "kind": 1},
                headers={**cls._headers},
            )
            if status := response.json().get("status"):
                if status != 0:
                    raise RequestException(f"获得cred失败：{response.json().get('messgae')}")
            return CRED(**response.json().get("data"))

    @classmethod
    async def refresh_token(cls, cred: str) -> str:
        async with httpx.AsyncClient() as client:
            refresh_url = "https://zonai.skland.com/api/v1/auth/refresh"
            try:
                response = await client.get(
                    refresh_url,
                    headers={**cls._headers, "cred": cred},
                )
                response.raise_for_status()
                if status := response.json().get("status"):
                    if status != 0:
                        raise RequestException(f"刷新token失败：{response.json().get('message')}")
                token = response.json().get("data").get("token")
                return token
            except (httpx.HTTPStatusError, httpx.ConnectError) as e:
                raise RequestException(f"刷新token失败：{str(e)}")



from dataclasses import field, dataclass

from nonebot.compat import PYDANTIC_V2
from pydantic import HttpUrl, BaseModel, ConfigDict


@dataclass
class Topics:
    topic: str
    topic_id: str = field(init=False)

    _MAPPING = {"萨米": "rogue_3", "萨卡兹": "rogue_4"}

    def __post_init__(self):
        self.topic_id = self._MAPPING[self.topic]


class Dynamic(BaseModel):
    type: int
    url: HttpUrl
    videoId: str
    kind: int
    filename: str
    transcodeStatus: int


class Char(BaseModel):
    id: str
    rarity: int
    profession: str
    type: str
    upgradePhase: int
    evolvePhase: int
    level: int
    name: str


class Tag(BaseModel):
    name: str
    icon: HttpUrl
    description: str
    id: int


class Band(BaseModel):
    id: str
    name: str


class Totem(BaseModel):
    id: str
    count: int


class Record(BaseModel):
    id: str
    modeGrade: int
    mode: str
    success: int
    lastChars: list[Char]
    initChars: list[Char]
    troopChars: list[Char]
    gainRelicList: list
    cntCrossedZone: int
    cntArrivedNode: int
    cntBattleNormal: int
    cntBattleElite: int
    cntBattleBoss: int
    cntGainRelicItem: int
    cntRecruitUpgrade: int
    totemList: list[Totem]
    seed: str
    tagList: list[Tag]
    lastStage: str
    score: int
    band: Band
    startTs: str
    endTs: str
    endingText: str
    isCollect: bool


class Medal(BaseModel):
    count: int
    current: int


class RogueHistory(BaseModel):
    medal: Medal
    modeGrade: int
    mode: str
    score: int
    bpLevel: int
    chars: list[Char]
    tagList: list[Tag]
    records: list[Record]
    favourRecords: list

    if PYDANTIC_V2:
        model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
    else:

        class Config:
            extra = "allow"
            arbitrary_types_allowed = True



from .hire import Hire as Hire
from .base import Labor as Labor
from .power import Power as Power
from .control import Control as Control
from .meeting import Meeting as Meeting
from .trading import Trading as Trading
from .base import Furniture as Furniture
from .tired import TiredChar as TiredChar
from .training import Training as Training
from .dormitory import Dormitory as Dormitory
from .manufacture import Manufacture as Manufacture
from .manufacture import ManufactureFormulaInfo as ManufactureFormulaInfo



from pydantic import BaseModel

from .base import Bubble


class TiredChar(BaseModel):
    """疲劳干员"""

    charId: str
    ap: int
    lastApAddTime: int
    roomSlotId: str
    index: int
    bubble: Bubble
    workTime: int



from sqlalchemy import Text
from nonebot_plugin_orm import Model
from sqlalchemy.orm import Mapped, mapped_column


class User(Model):
    __tablename__ = "skland_user"

    id: Mapped[int] = mapped_column(primary_key=True)
    """User ID"""
    access_token: Mapped[str] = mapped_column(Text, nullable=True)
    """Skland Access Token"""
    cred: Mapped[str] = mapped_column(Text)
    """Skland Login Credential"""
    cred_token: Mapped[str] = mapped_column(Text)
    """Skland Login Credential Token"""
    user_id: Mapped[int] = mapped_column(Text, nullable=True)
    """Skland User ID"""


class Character(Model):
    __tablename__ = "skland_characters"

    id: Mapped[int] = mapped_column(primary_key=True)
    """Character ID"""
    uid: Mapped[int] = mapped_column(primary_key=True)
    """Character UID"""
    app_code: Mapped[str] = mapped_column(Text)
    """APP Code"""
    channel_master_id: Mapped[str] = mapped_column(Text)
    """Channel Master ID"""
    nickname: Mapped[str] = mapped_column(Text)
    """Character Nickname"""
    isdefault: Mapped[bool] = mapped_column(default=False)



import hmac
import json
import hashlib
from typing import Literal
from datetime import datetime
from urllib.parse import urlparse

import httpx
from nonebot import logger

from ..schemas import CRED, ArkCard, RogueHistory, ArkSignResponse
from ..exception import LoginException, RequestException, UnauthorizedException

base_url = "https://zonai.skland.com/api/v1"


class SklandAPI:
    _headers = {
        "User-Agent": ("Skland/1.32.1 (com.hypergryph.skland; build:103201004; Android 33; ) Okhttp/4.11.0"),
        "Accept-Encoding": "gzip",
        "Connection": "close",
    }

    _header_for_sign = {"platform": "", "timestamp": "", "dId": "", "vName": ""}

    @classmethod
    async def get_binding(cls, cred: CRED) -> list:
        """获取绑定的角色"""
        binding_url = f"{base_url}/game/player/binding"
        async with httpx.AsyncClient() as client:
            try:
                response = await client.get(
                    binding_url,
                    headers=cls.get_sign_header(cred, binding_url, method="get"),
                )
                if status := response.json().get("code"):
                    if status == 10000:
                        raise UnauthorizedException(f"获取绑定角色失败：{response.json().get('message')}")
                    elif status == 10002:
                        raise LoginException(f"获取绑定角色失败：{response.json().get('message')}")
                    if status != 0:
                        raise RequestException(f"获取绑定角色失败：{response.json().get('message')}")
                return response.json()["data"]["list"]
            except httpx.HTTPError as e:
                raise RequestException(f"获取绑定角色失败: {e}")

    @classmethod
    def get_sign_header(
        cls,
        cred: CRED,
        url: str,
        method: Literal["get", "post"],
        query_body: dict | None = None,
    ) -> dict:
        """获取带sign请求头"""
        timestamp = int(datetime.now().timestamp()) - 1
        header_ca = {**cls._header_for_sign, "timestamp": str(timestamp)}
        parsed_url = urlparse(url)
        query_params = json.dumps(query_body) if method == "post" else parsed_url.query
        header_ca_str = json.dumps(
            {**cls._header_for_sign, "timestamp": str(timestamp)},
            separators=(",", ":"),
        )
        secret = f"{parsed_url.path}{query_params}{timestamp}{header_ca_str}"
        hex_secret = hmac.new(cred.token.encode("utf-8"), secret.encode("utf-8"), hashlib.sha256).hexdigest()
        signature = hashlib.md5(hex_secret.encode("utf-8")).hexdigest()
        return {"cred": cred.cred, **cls._headers, "sign": signature, **header_ca}

    @classmethod
    async def ark_sign(cls, cred: CRED, uid: str, channel_master_id: str) -> ArkSignResponse:
        """进行明日方舟签到"""
        body = {"uid": uid, "gameId": channel_master_id}
        json_body = json.dumps(body, ensure_ascii=False, separators=(", ", ": "), allow_nan=False)
        sign_url = f"{base_url}/game/attendance"
        headers = cls.get_sign_header(
            cred,
            sign_url,
            method="post",
            query_body=body,
        )
        async with httpx.AsyncClient() as client:
            try:
                response = await client.post(
                    sign_url,
                    headers={**headers, "Content-Type": "application/json"},
                    content=json_body,
                )
                logger.debug(f"签到回复：{response.json()}")
                if status := response.json().get("code"):
                    if status == 10000:
                        raise UnauthorizedException(f"角色 {uid} 签到失败：{response.json().get('message')}")
                    elif status == 10002:
                        raise LoginException(f"角色 {uid} 签到失败：{response.json().get('message')}")
                    elif status != 0:
                        raise RequestException(f"角色 {uid} 签到失败：{response.json().get('message')}")
            except httpx.HTTPError as e:
                raise RequestException(f"角色 {uid} 签到失败: {e}")
            return ArkSignResponse(**response.json()["data"])

    @classmethod
    async def get_user_ID(cls, cred: CRED) -> str:
        uid_url = f"{base_url}/user/teenager"
        async with httpx.AsyncClient() as client:
            try:
                response = await client.get(
                    uid_url,
                    headers=cls.get_sign_header(cred, uid_url, method="get"),
                )
                if status := response.json().get("code"):
                    if status == 10000:
                        raise UnauthorizedException(f"获取账号 userId 失败：{response.json().get('message')}")
                    elif status == 10002:
                        raise LoginException(f"获取账号 userId 失败：{response.json().get('message')}")
                    if status != 0:
                        raise RequestException(f"获取账号 userId 失败：{response.json().get('message')}")
                return response.json()["data"]["teenager"]["userId"]
            except httpx.HTTPError as e:
                raise RequestException(f"获取账号 userId 失败: {e}")

    @classmethod
    async def ark_card(cls, cred: CRED, uid: str) -> ArkCard:
        """获取明日方舟角色信息"""
        game_info_url = f"{base_url}/game/player/info?uid={uid}"
        async with httpx.AsyncClient() as client:
            try:
                response = await client.get(
                    game_info_url,
                    headers=cls.get_sign_header(cred, game_info_url, method="get"),
                )
                if status := response.json().get("code"):
                    if status == 10000:
                        raise UnauthorizedException(f"获取账号 game_info 失败：{response.json().get('message')}")
                    elif status == 10002:
                        raise LoginException(f"获取账号 game_info 失败：{response.json().get('message')}")
                    if status != 0:
                        raise RequestException(f"获取账号 game_info 失败：{response.json().get('message')}")
                return ArkCard(**response.json()["data"])
            except httpx.HTTPError as e:
                raise RequestException(f"获取账号 userId 失败: {e}")

    @classmethod
    async def get_rogue(cls, cred: CRED, uid: str, topic_id: str) -> RogueHistory:
        """获取肉鸽数据"""
        rogue_url = f"{base_url}/game/arknights/rogue?uid={uid}&targetUserId={cred.userId}&topicId={topic_id}"
        async with httpx.AsyncClient() as client:
            try:
                response = await client.get(
                    rogue_url,
                    headers=cls.get_sign_header(cred, rogue_url, method="get"),
                )
                logger.debug(f"肉鸽数据：{response.json()}")
                if status := response.json().get("code"):
                    if status == 10000:
                        raise UnauthorizedException(f"获取肉鸽数据失败：{response.json().get('message')}")
                    elif status == 10002:
                        raise LoginException(f"获取肉鸽数据失败：{response.json().get('message')}")
                    if status != 0:
                        raise RequestException(f"获取肉鸽数据失败：{response.json().get('message')}")
                return RogueHistory(**response.json()["data"]["history"])
            except httpx.HTTPError as e:
                raise RequestException(f"获取肉鸽数据失败: {e}")



from nonebot import require
from nonebot.plugin import PluginMetadata, inherit_supported_adapters

require("nonebot_plugin_orm")
require("nonebot_plugin_user")
require("nonebot_plugin_alconna")
require("nonebot_plugin_localstore")
require("nonebot_plugin_htmlrender")
from nonebot_plugin_orm import async_scoped_session
from nonebot_plugin_user import UserSession, get_user
from nonebot_plugin_alconna import (
    At,
    Args,
    Field,
    Match,
    Option,
    Alconna,
    Arparma,
    MsgTarget,
    Subcommand,
    UniMessage,
    CommandMeta,
    on_alconna,
)

from .model import User
from . import hook as hook
from .config import Config
from .render import render_ark_card
from .exception import RequestException
from .api import SklandAPI, SklandLoginAPI
from .schemas import CRED, Topics, ArkSignResponse
from .utils import get_characters_and_bind, refresh_cred_token_if_needed, refresh_access_token_if_needed
from .db_handler import get_arknights_characters, get_arknights_character_by_uid, get_default_arknights_character

__plugin_meta__ = PluginMetadata(
    name="森空岛",
    description="通过森空岛查询游戏数据",
    usage="/skland",
    config=Config,
    type="application",
    homepage="https://github.com/FrostN0v0/nonebot-plugin-skland",
    supported_adapters=inherit_supported_adapters("nonebot_plugin_alconna", "nonebot_plugin_user"),
    extra={
        "author": "FrostN0v0 <1614591760@qq.com>",
        "version": "0.1.4",
    },
)

skland = on_alconna(
    Alconna(
        "skland",
        Args["target?#目标", At | int],
        Subcommand(
            "-b|--bind|bind",
            Args["token", str, Field(completion=lambda: "请输入 token 或 cred 完成绑定")],
            Option("-u|--update|update", help_text="更新绑定的 token 或 cred"),
            help_text="绑定森空岛账号",
        ),
        Subcommand(
            "arksign",
            Option(
                "-u|--uid|uid",
                Args["uid", str, Field(completion=lambda: "请输入指定绑定角色uid")],
                help_text="指定绑定角色uid进行签到",
            ),
            Option("--all", help_text="签到所有绑定角色"),
            help_text="明日方舟签到",
        ),
        Subcommand("char", Option("-u|--update|update"), help_text="更新绑定角色信息"),
        Subcommand(
            "rogue",
            Args["target?#目标", At | int],
            Option(
                "-t|--topic|topic",
                Args["topic_name?#主题", ["萨米", "萨卡兹"], Field(completion=lambda: "请输入指定topic_id")],
                help_text="指定主题进行肉鸽战绩查询",
            ),
            help_text="肉鸽战绩查询",
        ),
        meta=CommandMeta(
            description=__plugin_meta__.description,
            usage=__plugin_meta__.usage,
            example="/skland",
        ),
    ),
    comp_config={"lite": True},
    skip_for_unmatch=False,
    block=True,
    use_cmd_start=True,
)

skland.shortcut("森空岛绑定", {"command": "skland bind", "fuzzy": True, "prefix": True})
skland.shortcut("明日方舟签到", {"command": "skland arksign --all", "fuzzy": True, "prefix": True})
skland.shortcut("萨卡兹肉鸽", {"command": "skland rogue --topic 萨卡兹", "fuzzy": True, "prefix": True})
skland.shortcut("萨米肉鸽", {"command": "skland rogue --topic 萨米", "fuzzy": True, "prefix": True})
skland.shortcut("角色更新", {"command": "skland char update", "fuzzy": False, "prefix": True})


@skland.assign("$main")
async def _(session: async_scoped_session, user_session: UserSession, target: Match[At | int]):
    @refresh_cred_token_if_needed
    @refresh_access_token_if_needed
    async def get_character_info(user: User, uid: str):
        return await SklandAPI.ark_card(CRED(cred=user.cred, token=user.cred_token), uid)

    if target.available:
        target_platform_id = target.result.target if isinstance(target.result, At) else target.result
        target_id = (await get_user(user_session.platform, str(target_platform_id))).id
    else:
        target_id = user_session.user_id

    user = await session.get(User, target_id)
    if not user:
        await UniMessage("未绑定 skland 账号").finish(at_sender=True)
    ark_characters = await get_default_arknights_character(user, session)
    if not ark_characters:
        await UniMessage("未绑定 arknights 账号").finish(at_sender=True)

    info = await get_character_info(user, str(ark_characters.uid))
    image = await render_ark_card(info)
    await UniMessage.image(raw=image).send()
    await session.commit()


@skland.assign("bind")
async def _(
    token: Match[str],
    result: Arparma,
    user_session: UserSession,
    msg_target: MsgTarget,
    session: async_scoped_session,
):
    """绑定森空岛账号"""
    if user := await session.get(User, user_session.user_id):
        if result.find("bind.update"):
            if len(token.result) == 24:
                grant_code = await SklandLoginAPI.get_grant_code(token.result)
                cred = await SklandLoginAPI.get_cred(grant_code)
                user.access_token = token.result
                user.cred = cred.cred
                user.cred_token = cred.token
            elif len(token.result) == 32:
                cred_token = await SklandLoginAPI.refresh_token(token.result)
                user.cred = token.result
                user.cred_token = cred_token
            else:
                await UniMessage("token 或 cred 错误,请检查格式").finish(at_sender=True)
            await get_characters_and_bind(user, session)
            await UniMessage("更新成功").finish(at_sender=True)
        await UniMessage("已绑定过 skland 账号").finish(at_sender=True)

    if not msg_target.private:
        await UniMessage("绑定指令只允许在私聊中使用").finish(at_sender=True)

    if token.available:
        try:
            if len(token.result) == 24:
                grant_code = await SklandLoginAPI.get_grant_code(token.result)
                cred = await SklandLoginAPI.get_cred(grant_code)
                user = User(
                    access_token=token.result,
                    cred=cred.cred,
                    cred_token=cred.token,
                    id=user_session.user_id,
                    user_id=cred.userId,
                )
            elif len(token.result) == 32:
                cred_token = await SklandLoginAPI.refresh_token(token.result)
                user_id = await SklandAPI.get_user_ID(CRED(cred=token.result, token=cred_token))
                user = User(
                    cred=token.result,
                    cred_token=cred_token,
                    id=user_session.user_id,
                    user_id=user_id,
                )
            else:
                await UniMessage("token 或 cred 错误,请检查格式").finish(at_sender=True)
            session.add(user)
            await get_characters_and_bind(user, session)
            await UniMessage("绑定成功").finish(at_sender=True)
        except RequestException as e:
            await UniMessage(f"绑定失败,错误信息:{e}").finish(at_sender=True)


@skland.assign("arksign")
async def _(user_session: UserSession, session: async_scoped_session, uid: Match[str], result: Arparma):
    """明日方舟森空岛签到"""

    @refresh_cred_token_if_needed
    @refresh_access_token_if_needed
    async def sign_in(user: User, uid: str, channel_master_id: str):
        """执行签到逻辑"""
        cred = CRED(cred=user.cred, token=user.cred_token)
        return await SklandAPI.ark_sign(cred, uid, channel_master_id=channel_master_id)

    user = await session.get(User, user_session.user_id)
    if not user:
        await UniMessage("未绑定 skland 账号").finish(at_sender=True)

    sign_result: dict[str, ArkSignResponse] = {}
    if uid.available:
        character = await get_arknights_character_by_uid(user, uid.result, session)
        sign_result[character.nickname] = await sign_in(user, uid.result, character.channel_master_id)
    else:
        if result.find("arksign.all"):
            characters = await get_arknights_characters(user, session)
            for character in characters:
                sign_result[character.nickname] = await sign_in(user, str(character.uid), character.channel_master_id)
        else:
            character = await get_default_arknights_character(user, session)
            if not character:
                await UniMessage("未绑定 arknights 账号").finish(at_sender=True)

            sign_result[character.nickname] = await sign_in(user, str(character.uid), character.channel_master_id)

    if sign_result[character.nickname]:
        await UniMessage(
            "\n".join(
                f"角色: {nickname} 签到成功，获得了:\n"
                + "\n".join(f"{award.resource.name} x {award.count}" for award in sign.awards)
                for nickname, sign in sign_result.items()
            )
        ).send(at_sender=True)

    await session.commit()


@skland.assign("char.update")
async def _(user_session: UserSession, session: async_scoped_session):
    """更新森空岛角色信息"""

    @refresh_cred_token_if_needed
    @refresh_access_token_if_needed
    async def refresh_characters(user: User):
        await get_characters_and_bind(user, session)
        await UniMessage("更新成功").send(at_sender=True)

    if user := await session.get(User, user_session.user_id):
        await refresh_characters(user)


@skland.assign("rogue")
async def _(user_session: UserSession, session: async_scoped_session, result: Arparma, target: Match[At | int]):
    """获取明日方舟肉鸽战绩"""

    # Not Finished
    @refresh_cred_token_if_needed
    @refresh_access_token_if_needed
    async def get_rogue_info(user: User, uid: str, topic_id: str):
        return await SklandAPI.get_rogue(
            CRED(cred=user.cred, token=user.cred_token, userId=str(user.user_id)), uid, topic_id
        )

    if target.available:
        target_platform_id = target.result.target if isinstance(target.result, At) else target.result
        target_id = (await get_user(user_session.platform, str(target_platform_id))).id
    else:
        target_id = user_session.user_id

    user = await session.get(User, target_id)
    if not user:
        await UniMessage("未绑定 skland 账号").finish(at_sender=True)
    character = await get_default_arknights_character(user, session)
    if not character:
        await UniMessage("未绑定 arknights 账号").finish(at_sender=True)

    topic_id = Topics(str(result.query("rogue.topic.topic_name"))).topic_id if result.find("rogue.topic") else ""
    # TODO: 渲染肉鸽战绩卡片，完善指令逻辑
    rogue = await get_rogue_info(user, str(character.uid), topic_id)  # noqa: F841
    # await UniMessage(rogue.model_dump_json()).send()
    await UniMessage("功能开发中（救命!来画渲染模板").send()
    await session.commit()



from .cred import CRED as CRED
from .rogue import Topics as Topics
from .ark_card import ArkCard as ArkCard
from .rogue import RogueHistory as RogueHistory
from .ark_models import AssistChar as AssistChar
from .ark_sign import ArkSignResponse as ArkSignResponse



from pathlib import Path

from pydantic import Field, BaseModel
import nonebot_plugin_localstore as store
from nonebot.plugin import get_plugin_config

RES_DIR: Path = Path(__file__).parent / "resources"
TEMPLATES_DIR: Path = RES_DIR / "templates"
CACHE_DIR = store.get_plugin_cache_dir()


class ScopedConfig(BaseModel):
    github_proxy_url: str = ""
    """GitHub 代理 URL"""
    github_token: str = ""
    """GitHub Token"""
    check_res_update: bool = True
    """检查资源更新"""


class Config(BaseModel):
    skland: ScopedConfig = Field(default_factory=ScopedConfig)


config = get_plugin_config(Config).skland



from .skins import Skin as Skin
from .medal import Medal as Medal
from .tower import Tower as Tower
from .status import Status as Status
from .recruit import Recruit as Recruit
from .routine import Routine as Routine
from .chars import Character as Character
from .building import Building as Building
from .campaign import Campaign as Campaign
from .assist_chars import Equipment as Equipment
from .assist_chars import AssistChar as AssistChar
from .buildings import ManufactureFormulaInfo as ManufactureFormulaInfo



import math
from datetime import datetime

from pydantic import BaseModel


class Avatar(BaseModel):
    """角色头像信息"""

    type: str
    id: str
    url: str


class Secretary(BaseModel):
    """助理干员信息"""

    charId: str
    skinId: str


class AP(BaseModel):
    """理智"""

    current: int
    max: int
    lastApAddTime: int
    completeRecoveryTime: int

    @property
    def ap_now(self) -> int:
        """计算当前理智 ap_now ,并确保不超过最大理智值。"""
        current_time = datetime.now().timestamp()
        ap_now = self.max - math.ceil((self.completeRecoveryTime - current_time) / 360)
        ap_now = min(ap_now, self.max)

        return ap_now


class Exp(BaseModel):
    """经验值"""

    current: int
    max: int


class Status(BaseModel):
    """
    角色状态信息

    Attributes:
        uid : 角色 UID
        name : 角色名称
        level : 等级
        avatar : 头像信息
        registerTs : 注册时间戳
        secretary : 助理干员信息
        ap :理智信息
        lastOnlineTs : 角色最后在线时间戳
        exp : 经验值
    """

    uid: str
    name: str
    level: int
    avatar: Avatar
    registerTs: int
    mainStageProgress: str
    secretary: Secretary
    resume: str
    subscriptionEnd: int
    ap: AP
    storeTs: int
    lastOnlineTs: int
    charCnt: int
    furnitureCnt: int
    skinCnt: int
    exp: Exp



from pydantic import BaseModel

from .base import BaseCount


class Routine(BaseModel):
    """
    日/周常任务完成进度

    Attributes:
        daily : 日常任务进度
        weekly : 周常任务进度
    """

    daily: BaseCount
    weekly: BaseCount



"""first_revision

迁移 ID: 997049a57a3a
父迁移:
创建时间: 2025-02-26 09:22:45.310083

"""

from __future__ import annotations

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

revision: str = "997049a57a3a"
down_revision: str | Sequence[str] | None = None
branch_labels: str | Sequence[str] | None = ("nonebot_plugin_skland",)
depends_on: str | Sequence[str] | None = None


def upgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "skland_characters",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("uid", sa.Integer(), nullable=False),
        sa.Column("app_code", sa.Text(), nullable=False),
        sa.Column("channel_master_id", sa.Text(), nullable=False),
        sa.Column("nickname", sa.Text(), nullable=False),
        sa.Column("isdefault", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id", "uid", name=op.f("pk_skland_characters")),
        info={"bind_key": "nonebot_plugin_skland"},
    )
    op.create_table(
        "skland_user",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("access_token", sa.Text(), nullable=True),
        sa.Column("cred", sa.Text(), nullable=False),
        sa.Column("cred_token", sa.Text(), nullable=False),
        sa.Column("user_id", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_skland_user")),
        info={"bind_key": "nonebot_plugin_skland"},
    )
    # ### end Alembic commands ###


def downgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("skland_user")
    op.drop_table("skland_characters")
    # ### end Alembic commands ###



from typing import Literal, overload

from nonebot.compat import PYDANTIC_V2
from nonebot.compat import model_dump as model_dump

__all__ = (
    "model_dump",
    "model_validator",
)


if PYDANTIC_V2:
    from pydantic import model_validator as model_validator
else:
    from pydantic import root_validator

    @overload
    def model_validator(*, mode: Literal["before"]): ...

    @overload
    def model_validator(*, mode: Literal["after"]): ...

    def model_validator(*, mode: Literal["before", "after"]):
        return root_validator(pre=mode == "before", allow_reuse=True)  # type: ignore



from datetime import datetime

from pydantic import BaseModel

from .buildings import (
    Hire,
    Labor,
    Power,
    Control,
    Meeting,
    Trading,
    Training,
    Dormitory,
    Furniture,
    TiredChar,
    Manufacture,
)


class Building(BaseModel):
    """
    基建信息

    Attributes:
        tiredChars :  疲劳干员
        powers :  发电站
        manufactures :  制造站
        tradings :  交易站
        dormitories :  宿舍
        meeting :  会客室
        hire :  人力办公室
        training :  训练室
        labor :  无人机
        furniture :  家具
        control :  控制中枢
    """

    tiredChars: list[TiredChar]
    powers: list[Power]
    manufactures: list[Manufacture]
    tradings: list[Trading]
    dormitories: list[Dormitory]
    meeting: Meeting
    hire: Hire
    training: Training
    labor: Labor
    furniture: Furniture
    control: Control

    @property
    def rested_chars(self):
        """此处未计算基建技能影响，因此实际休息进度可能有差异"""
        rested_count = 0
        for dorm in self.dormitories:
            for char in dorm.chars:
                ap_gain_rate = 1.5 + dorm.level * 0.1 + 0.0004 * dorm.comfort * 100
                time_diff = datetime.now().timestamp() - char.lastApAddTime
                ap_now = min(char.ap + time_diff * ap_gain_rate, 8640000)
                if ap_now == 8640000:
                    rested_count += 1
        return rested_count

    @property
    def dorm_chars(self):
        dorm_char_count = 0
        for dorm in self.dormitories:
            dorm_char_count += len(dorm.chars)
        return dorm_char_count

    @property
    def trading_stock(self):
        """获取交易站库存"""
        stock_count = 0
        for trading in self.tradings:
            if trading.completeWorkTime >= datetime.now().timestamp():
                stock_count += 1
            stock_count += len(trading.stock)
        return stock_count

    @property
    def trading_stock_limit(self):
        """获取交易站库存上限"""
        stock_limit = 0
        for trading in self.tradings:
            stock_limit += trading.stockLimit
        return stock_limit



from datetime import datetime

from pydantic import BaseModel

from .ark_models import (
    Skin,
    Medal,
    Tower,
    Status,
    Recruit,
    Routine,
    Building,
    Campaign,
    Character,
    Equipment,
    AssistChar,
    ManufactureFormulaInfo,
)


class CharInfo(BaseModel):
    id: str
    name: str


class ArkCard(BaseModel):
    status: Status
    medal: Medal
    assistChars: list[AssistChar]
    chars: list[Character]
    skins: list[Skin]
    recruit: list[Recruit]
    campaign: Campaign
    tower: Tower
    routine: Routine
    building: Building
    equipmentInfoMap: dict[str, Equipment]
    manufactureFormulaInfoMap: dict[str, ManufactureFormulaInfo]
    charInfoMap: dict[str, CharInfo]

    @property
    def recruit_finished(self) -> int:
        return len([recruit for recruit in self.recruit if recruit.state == 1])

    @property
    def recruit_complete_time(self) -> str:
        from ..render import format_timestamp

        finish_ts = max([recruit.finishTs for recruit in self.recruit])
        if finish_ts == -1:
            return "招募已全部完成"
        format_time = format_timestamp(finish_ts - datetime.now().timestamp())
        return f"{format_time}后全部完成"



from nonebot import logger
from nonebot_plugin_alconna import UniMessage
from nonebot_plugin_orm import async_scoped_session

from .schemas import CRED
from .model import User, Character
from .db_handler import delete_characters
from .api import SklandAPI, SklandLoginAPI
from .exception import LoginException, RequestException, UnauthorizedException


async def get_characters_and_bind(user: User, session: async_scoped_session):
    await delete_characters(user, session)

    cred = CRED(cred=user.cred, token=user.cred_token)
    binding_app_list = await SklandAPI.get_binding(cred)
    for app in binding_app_list:
        for character in app["bindingList"]:
            character_model = Character(
                id=user.id,
                uid=character["uid"],
                nickname=character["nickName"],
                app_code=app["appCode"],
                channel_master_id=character["channelMasterId"],
                isdefault=character["isDefault"],
            )
            if len(app["bindingList"]) == 1:
                character_model.isdefault = True
            session.add(character_model)
    await session.commit()


def refresh_access_token_if_needed(func):
    """装饰器：如果 access_token 失效，刷新后重试"""

    async def wrapper(user: User, *args, **kwargs):
        try:
            return await func(user, *args, **kwargs)
        except LoginException:
            if not user.access_token:
                await UniMessage("cred失效，用户没有绑定token，无法自动刷新cred").send(at_sender=True)

            try:
                grant_code = await SklandLoginAPI.get_grant_code(user.access_token)
                new_cred = await SklandLoginAPI.get_cred(grant_code)
                user.cred, user.cred_token = new_cred.cred, new_cred.token
                logger.info("access_token 失效，已自动刷新")
                return await func(user, *args, **kwargs)
            except (RequestException, LoginException, UnauthorizedException) as e:
                await UniMessage(f"接口请求失败,错误信息:{e}").send(at_sender=True)
        except RequestException as e:
            await UniMessage(f"接口请求失败,错误信息:{e}").send(at_sender=True)

    return wrapper


def refresh_cred_token_if_needed(func):
    """装饰器：如果 cred_token 失效，刷新后重试"""

    async def wrapper(user: User, *args, **kwargs):
        try:
            return await func(user, *args, **kwargs)
        except UnauthorizedException:
            try:
                new_token = await SklandLoginAPI.refresh_token(user.cred)
                user.cred_token = new_token
                logger.info("cred_token 失效，已自动刷新")
                return await func(user, *args, **kwargs)
            except (RequestException, LoginException, UnauthorizedException) as e:
                await UniMessage(f"接口请求失败,错误信息:{e}").send(at_sender=True)
        except RequestException as e:
            await UniMessage(f"接口请求失败,错误信息:{e}").send(at_sender=True)

    return wrapper



from pydantic import BaseModel


class Trainee(BaseModel):
    charId: str
    targetSkill: int
    ap: int
    lastApAddTime: int


class Trainer(BaseModel):
    charId: str
    ap: int
    lastApAddTime: int


class Training(BaseModel):
    slotId: str
    level: int
    trainee: Trainee | None = None
    trainer: Trainer | None = None
    remainPoint: float
    speed: float
    lastUpdateTime: int
    remainSecs: int
    slotState: int

    @property
    def training_state(self) -> str:
        if self.trainee:
            if self.trainee.targetSkill == -1:
                return "空闲中"
            else:
                match self.trainee.targetSkill:
                    case -1:
                        return "空闲中"
                    case 0:
                        return "1技能"
                    case 1:
                        return "2技能"
                    case 2:
                        return "3技能"
        return "空闲中"



from pydantic import BaseModel

from .base import BaseCount


class TowerRecord(BaseModel):
    """
    保全派驻记录

    Attributes:
        towerId : 保全派驻 ID
        best : 最高进度
    """

    towerId: str
    best: int


class TowerReward(BaseModel):
    """保全派驻奖励进度"""

    higherItem: BaseCount
    lowerItem: BaseCount
    termTs: int


class Tower(BaseModel):
    """保全派驻信息"""

    records: list[TowerRecord]
    reward: TowerReward



from pydantic import BaseModel

from .base import BuildingChar


class Manufacture(BaseModel):
    """制造站"""

    slotId: str
    level: int
    chars: list[BuildingChar]
    completeWorkTime: int
    lastUpdateTime: int
    formulaId: str
    capacity: int
    weight: int
    complete: int
    remain: int
    speed: float


class ManufactureFormulaInfo(BaseModel):
    """制造站配方"""

    id: str
    itemId: str
    weight: int
    costPoint: int



from datetime import datetime

from pydantic import BaseModel

from .base import BuildingChar


class Hire(BaseModel):
    """人事办公室"""

    slotId: str
    level: int
    chars: list[BuildingChar]
    state: int
    refreshCount: int
    completeWorkTime: int
    slotState: int

    @property
    def refresh_complete_time(self) -> str:
        from ....render import format_timestamp

        if self.refreshCount == 3:
            return "可进行公开招募标签刷新"
        format_time = format_timestamp(self.completeWorkTime - datetime.now().timestamp())
        return f"{format_time}后次数回满"



from sqlalchemy import delete, select
from nonebot_plugin_orm import async_scoped_session

from .model import User, Character


async def get_arknights_characters(user: User, session: async_scoped_session) -> list[Character]:
    characters = (
        (
            await session.execute(
                select(Character).where(Character.id == user.id).where(Character.app_code == "arknights")
            )
        )
        .scalars()
        .all()
    )
    return list(characters)


async def get_default_arknights_character(user: User, session: async_scoped_session) -> Character:
    character = (
        await session.execute(
            select(Character).where(
                Character.id == user.id,
                Character.isdefault,
                Character.app_code == "arknights",
            )
        )
    ).scalar_one()
    return character


async def get_arknights_character_by_uid(user: User, uid: str, session: async_scoped_session) -> Character:
    character = (
        await session.execute(
            select(Character).where(
                Character.id == user.id,
                Character.uid == int(uid),
                Character.app_code == "arknights",
            )
        )
    ).scalar_one()
    return character


async def delete_characters(user: User, session: async_scoped_session):
    await session.execute(delete(Character).where(Character.id == user.id))



from pydantic import BaseModel

from .base import BuildingChar


class Clue(BaseModel):
    """线索信息"""

    own: int
    received: int
    dailyReward: bool
    needReceive: int
    board: list[str]
    sharing: bool
    shareCompleteTime: int


class Meeting(BaseModel):
    """会客厅"""

    slotId: str
    level: int
    chars: list[BuildingChar]
    clue: Clue
    lastUpdateTime: int
    completeWorkTime: int



import asyncio
from pathlib import Path
from itertools import islice
from datetime import datetime
from urllib.parse import quote
from collections.abc import Iterable

from nonebot import logger
from rich.text import Text
from rich.panel import Panel
from rich.table import Table
from httpx import AsyncClient
from pydantic import BaseModel
from rich.progress import (
    Task,
    TaskID,
    Progress,
    BarColumn,
    TextColumn,
    DownloadColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from .compat import model_validator
from .config import CACHE_DIR, config


class File(BaseModel):
    name: str
    download_url: str

    @model_validator(mode="before")
    def modify_download_url(cls, values):
        values["download_url"] = quote(values["download_url"], safe="/:")
        if config.github_proxy_url:
            values["download_url"] = f"{config.github_proxy_url}{values['download_url']}"
            return values
        return values


class DownloadProgress(Progress):
    """下载进度条"""

    STATUS_DL = TextColumn("[blue]Downloading...")
    STATUS_FIN = TextColumn("[green]Complete!")
    STATUS_ROW = (
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%", justify="center"),
        TimeRemainingColumn(compact=True),
    )
    PROG_ROW = (DownloadColumn(binary_units=True), BarColumn(), TransferSpeedColumn())

    MAX_VISIBLE_TASKS = 10

    def make_tasks_table(self, tasks: Iterable[Task]) -> Table:
        table = Table.grid(padding=(0, 1), expand=self.expand)
        tasks_table = Table.grid(padding=(0, 1), expand=self.expand)
        all_tasks_finished = True
        visible_tasks = list(islice((task for task in tasks if task.visible), self.MAX_VISIBLE_TASKS))

        for task in visible_tasks:
            status = self.STATUS_FIN if task.finished else self.STATUS_DL
            itable = Table.grid(padding=(0, 1), expand=self.expand)
            filename_column = Text(f"{task.fields['filename']}")
            itable.add_row(filename_column, *(column(task) for column in [status, *self.STATUS_ROW]))
            itable.add_row(*(column(task) for column in self.PROG_ROW))
            tasks_table.add_row(itable)
            if not task.finished:
                all_tasks_finished = False

        if any(not task.finished for task in tasks):
            all_tasks_finished = False

        if all_tasks_finished:
            return table
        else:
            table.add_row(Panel(tasks_table, title="Downloading Files", title_align="left", padding=(1, 2)))

        return table


class GameResourceDownloader:
    """游戏数据下载"""

    DOWNLOAD_COUNT: int = 0
    DOWNLOAD_TIME: datetime
    SEMAPHORE = asyncio.Semaphore(100)
    RAW_BASE_URL = "https://raw.githubusercontent.com/{owner}/{repo}/{branch}/"
    VERSION_URL = "https://raw.githubusercontent.com/yuanyan3060/ArknightsGameResource/refs/heads/main/version"
    BASE_URL = "https://api.github.com/repos/{owner}/{repo}/git/trees/{branch}?recursive=1"

    @classmethod
    async def check_update(cls) -> bool | str:
        """检查更新"""
        url = config.github_proxy_url + cls.VERSION_URL if config.github_proxy_url else cls.VERSION_URL
        async with AsyncClient() as client:
            response = await client.get(url)
            response.raise_for_status()
            origin_version = response.content.decode()
            version_file = CACHE_DIR.joinpath("version")
            if not version_file.exists():
                version_file.write_text(origin_version, encoding="utf-8")
                return True
            local_version = version_file.read_text(encoding="utf-8").strip()
            if origin_version != local_version:
                return origin_version
            return False

    @classmethod
    def update_version_file(cls, version: str):
        """更新版本文件"""
        version_file = CACHE_DIR.joinpath("version")
        version_file.write_text(version, encoding="utf-8")

    @classmethod
    async def fetch_file_list(cls, url: str, dl_url: str, route: str) -> list[File]:
        """获取 GitHub 仓库下的所有文件，并返回可下载的 URL"""
        headers = {}
        if config.github_token:
            headers = {"Authorization": f"{config.github_token}"}
        async with AsyncClient() as client:
            response = await client.get(url, headers=headers)
            response.raise_for_status()
            data = response.json()
            route = route.rstrip("/") + "/"
            files = [
                File(name=item["path"].split("/")[-1], download_url=f"{dl_url}{item['path']}")
                for item in data.get("tree", [])
                if item["type"] == "blob" and item["path"].startswith(route)
            ]
            return files

    @classmethod
    async def download_all(cls, owner: str, repo: str, route: str, branch: str = "main"):
        """并行下载 GitHub 目录下的所有文件"""
        cls.download_count = 0
        cls.download_time = datetime.now()
        url = cls.BASE_URL.format(owner=owner, repo=repo, branch=branch)
        dl_url = cls.RAW_BASE_URL.format(owner=owner, repo=repo, branch=branch)
        files = await cls.fetch_file_list(url=url, dl_url=dl_url, route=route)
        save_path = CACHE_DIR / route
        save_path.mkdir(parents=True, exist_ok=True)

        async with AsyncClient() as client:
            with DownloadProgress(
                "[cyan]{task.fields[filename]}",
                BarColumn(),
                DownloadColumn(),
                TransferSpeedColumn(),
                TimeRemainingColumn(),
            ) as progress:

                async def worker(file: File):
                    """每个文件下载任务"""
                    if (save_path / file.name).exists():
                        return
                    async with cls.SEMAPHORE:
                        task_id = progress.add_task("Downloading", filename=file.name, total=0)
                        await cls.download_file(client, file, save_path, progress, task_id=task_id, timeout=300)
                        progress.remove_task(task_id)
                        cls.download_count += 1

                await asyncio.gather(*(worker(file) for file in files))
        time_consumed = datetime.now() - cls.download_time
        if cls.download_count == 0:
            logger.info(f"资源 {route} 无新增文件")
        else:
            logger.success(f"资源 {route} 下载完成，共下载 {cls.download_count} 个文件,耗时 {time_consumed}")

    @classmethod
    async def download_file(
        cls,
        client: AsyncClient,
        file: File,
        save_path: Path,
        progress: Progress,
        *,
        task_id: TaskID,
        **kwargs,
    ):
        """下载单个文件"""

        file_path = save_path / file.name
        async with client.stream("GET", file.download_url, **kwargs) as response:
            file_size = int(response.headers.get("Content-Length", 0))
            progress.update(task_id, total=file_size)

            with file_path.open("wb") as f:
                async for data in response.aiter_bytes(1024):
                    f.write(data)
                    progress.update(task_id, advance=len(data))



from pydantic import BaseModel


class Recruit(BaseModel):
    """
    公招信息

    Attributes:
        startTs : 开始时间戳
        finishTs : 结束时间戳
        state : 状态
    """

    startTs: int
    finishTs: int
    state: int



from pydantic import BaseModel

from .base import Equip
from ...config import RES_DIR, CACHE_DIR


class AssistChar(BaseModel):
    """
    助战干员

    Attributes:
        charId : 干员 ID
        skinId : 皮肤 ID
        level : 等级
        evolvePhase : 升级阶段
        potentialRank : 潜能等级
        skillId : 技能 ID
        mainSkillLvl : 主技能等级
        specializeLevel : 专精等级
        equip : 装备技能
    """

    charId: str
    skinId: str
    level: int
    evolvePhase: int
    potentialRank: int
    skillId: str
    mainSkillLvl: int
    specializeLevel: int
    equip: Equip | None = None
    uniequip: str | None = None

    @property
    def portrait(self) -> str:
        for symbol in ["@", "#"]:
            if symbol in self.skinId:
                portrait_id = self.skinId.replace(symbol, "_", 1)
                break
        img_path = CACHE_DIR / "portrait" / f"{portrait_id}.png"
        return img_path.as_uri()

    @property
    def potential(self) -> str:
        img_path = RES_DIR / "images" / "ark_card" / "potential" / f"potential_{self.potentialRank}.png"
        return img_path.as_uri()

    @property
    def skill(self) -> str:
        img_path = CACHE_DIR / "skill" / f"skill_icon_{self.skillId}.png"
        return img_path.as_uri()

    @property
    def evolve(self) -> str:
        img_path = RES_DIR / "images" / "ark_card" / "elite" / f"elite_{self.evolvePhase}.png"
        return img_path.as_uri()


class Equipment(BaseModel):
    id: str
    name: str
    typeIcon: str



from pydantic import BaseModel

from .base import BaseCount


class CampaignRecord(BaseModel):
    """
    剿灭记录

    Attributes:
        campaignId : 剿灭 ID
        maxKills : 最大击杀数
    """

    campaignId: str
    maxKills: int


class Campaign(BaseModel):
    """剿灭作战信息"""

    records: list[CampaignRecord]
    reward: BaseCount



from datetime import datetime, timedelta

from nonebot_plugin_htmlrender import template_to_pic

from .schemas import ArkCard
from .config import RES_DIR, TEMPLATES_DIR


async def render_ark_card(props: ArkCard) -> bytes:
    register_time = datetime.fromtimestamp(props.status.registerTs).strftime("%Y-%m-%d")
    main_progress = props.status.mainStageProgress if props.status.mainStageProgress else "全部完成"
    for char in props.assistChars:
        if char.equip:
            if char.equip.id in props.equipmentInfoMap.keys():
                equip_id = props.equipmentInfoMap[char.equip.id].typeIcon
                uniequip_path = RES_DIR / "images" / "ark_card" / "uniequip" / f"{equip_id}.png"
                char.uniequip = uniequip_path.as_uri()
        else:
            uniequip_path = RES_DIR / "images" / "ark_card" / "uniequip" / "original.png"
            char.uniequip = uniequip_path.as_uri()
    stoke_max = 0
    stoke_count = 0
    for manufacture in props.building.manufactures:
        if manufacture.formulaId in props.manufactureFormulaInfoMap.keys():
            formula_weight = props.manufactureFormulaInfoMap[manufacture.formulaId].weight
            stoke_max += int(manufacture.capacity / formula_weight)
            elapsed_time = datetime.now().timestamp() - manufacture.lastUpdateTime
            cost_time = props.manufactureFormulaInfoMap[manufacture.formulaId].costPoint / manufacture.speed
            additional_complete = round(elapsed_time / cost_time)
            if datetime.now().timestamp() >= manufacture.completeWorkTime:
                stoke_count += manufacture.capacity // formula_weight
            else:
                to_be_processed = (manufacture.completeWorkTime - manufacture.lastUpdateTime) / (
                    cost_time / manufacture.speed
                )
                has_processed = to_be_processed - int(to_be_processed)
                additional_complete = (elapsed_time - has_processed * cost_time) / cost_time
                stoke_count += manufacture.complete + int(additional_complete) + 1

    trainee_char_id = props.building.training.trainee.charId if props.building.training.trainee else ""
    if trainee_char_id in props.charInfoMap.keys():
        trainee_char_name = props.charInfoMap[trainee_char_id].name
    else:
        trainee_char_name = ""

    ap_recovery_time = format_timestamp(props.status.ap.completeRecoveryTime - datetime.now().timestamp())
    if props.status.ap.completeRecoveryTime > datetime.now().timestamp():
        ap_recovery = f"{ap_recovery_time}后全部恢复"
    else:
        ap_recovery = "已全部恢复"

    return await template_to_pic(
        template_path=str(TEMPLATES_DIR),
        template_name="ark_card.html.jinja2",
        templates={
            "Dr_name": props.status.name,
            "Dr_level": props.status.level,
            "Dr_avatar": props.status.avatar.url,
            "register_time": register_time,
            "main_progress": main_progress,
            "employed_chars": len(props.chars),
            "skins": len(props.skins),
            "furniture": props.building.furniture.total,
            "medals": props.medal.total,
            "assist_chars": props.assistChars,
            "labor": props.building.labor,
            "rested": props.building.rested_chars,
            "dorm_chars": props.building.dorm_chars,
            "trading_stoke": props.building.trading_stock,
            "trading_stoke_limit": props.building.trading_stock_limit,
            "manu_stoke_max": stoke_max,
            "manu_stoke_count": stoke_count,
            "tired": len(props.building.tiredChars),
            "clue": len(props.building.meeting.clue.board),
            "ap": props.status.ap,
            "ap_recovery": ap_recovery,
            "recruit_finished": props.recruit_finished,
            "recruit_max": len(props.recruit),
            "recruit_complete_time": props.recruit_complete_time,
            "refresh_count": props.building.hire.refreshCount,
            "refresh_complete_time": props.building.hire.refresh_complete_time,
            "campaign": props.campaign,
            "weekly_refresh": time_to_next_monday_4am(),
            "daily_refresh": time_to_next_4am(),
            "routine": props.routine,
            "tower": props.tower.reward,
            "tower_refresh": format_timestamp(props.tower.reward.termTs - datetime.now().timestamp()),
            "training_char": trainee_char_name,
            "training_state": props.building.training.training_state,
            "training_complete_time": format_timestamp(props.building.training.remainSecs),
        },
        pages={
            "viewport": {"width": 706, "height": 1160},
            "base_url": f"file://{TEMPLATES_DIR}",
        },
    )


def format_timestamp(timestamp: float) -> str:
    delta = timedelta(seconds=timestamp)
    days = delta.days
    hours, remainder = divmod(delta.seconds, 3600)
    minutes = remainder // 60

    if days > 0:
        return f"{days}天{hours}小时{minutes}分钟"
    elif hours > 0:
        return f"{hours}小时{minutes}分钟"
    else:
        return f"{minutes}分钟"


def time_to_next_monday_4am():
    now = datetime.now()
    days_until_monday = (7 - now.weekday()) % 7
    next_monday = now + timedelta(days=days_until_monday)
    next_monday_4am = next_monday.replace(hour=4, minute=0, second=0, microsecond=0)
    if now > next_monday_4am:
        next_monday_4am += timedelta(weeks=1)

    return format_timestamp(next_monday_4am.timestamp() - now.timestamp())


def time_to_next_4am():
    now = datetime.now()
    next_4am = now.replace(hour=4, minute=0, second=0, microsecond=0)
    if now > next_4am:
        next_4am += timedelta(days=1)
    return format_timestamp(next_4am.timestamp() - now.timestamp())



from pydantic import BaseModel


class BaseCount(BaseModel):
    """
    获取/完成进度

    Attributes:
        current (int): 当前值。
        total (int): 总值/上限。
    """

    current: int
    total: int


class Equip(BaseModel):
    """
    干员装备技能

    Attributes:
        id : 技能 ID
        level : 等级
    """

    id: str
    level: int
    locked: bool



from pydantic import BaseModel

from .base import BuildingChar


class DeliveryItem(BaseModel):
    id: str
    count: int
    type: str


class Gain(BaseModel):
    id: str
    count: int
    type: str


class StockItem(BaseModel):
    """储存订单"""

    instId: int
    type: str
    delivery: list[DeliveryItem]
    gain: Gain
    isViolated: bool


class Trading(BaseModel):
    """贸易站"""

    slotId: str
    level: int
    chars: list[BuildingChar]
    completeWorkTime: int
    lastUpdateTime: int
    strategy: str
    stock: list[StockItem]
    stockLimit: int



from pydantic import BaseModel


class MedalLayout(BaseModel):
    """
    蚀刻章布局

    Attributes:
        id (str): 奖章的唯一标识符
        pos (List[int]): 奖章的坐标位置,包含两个整数值(x, y)
    """

    id: str
    pos: list[int]


class Medal(BaseModel):
    """
    佩戴蚀刻章信息。

    Attributes:
        type (str): 佩戴蚀刻章类型，自定义或者套装
        template (str): 蚀刻章模板
        templateMedalList (List): 模板蚀刻章列表
        customMedalLayout (List[MedalLayout]): 蚀刻章的自定义布局
        total (int): 拥有的蚀刻章总数
    """

    type: str
    template: str
    templateMedalList: list[MedalLayout]
    customMedalLayout: list[MedalLayout]
    total: int
